import os
import redis
import socket
from time import sleep
from threading import Thread
from .base import BaseMode

DEFAULT_REDIS_ADDR = "redis_streaming"
DEFAULT_REDIS_PORT = 6379
DEFAULT_LISTEN_ADDR = "0.0.0.0"
DEFAULT_LISTEN_PORT = 5556
DEFAULT_DETRIGGER_DELAY = 5
TRIGGER_ID = int(os.getenv("TRIGGER_ID", 0))
NODE_ADDR = os.getenv("NODE_ADDR", "")


class NetowrkTriggerMode(BaseMode):
    def __init__(self, config, trigger_responses):
        super().__init__(config, trigger_responses)
        self.name = "network_trigger"
        self.trigger_id = config["trigger_id"] if "trigger_id" in config else TRIGGER_ID
        self.listen_addr = config["listen_addr"] if "listen_addr" in config else DEFAULT_LISTEN_ADDR
        self.listen_port = config["listen_port"] if "listen_port" in config else DEFAULT_LISTEN_PORT
        self.detrigger_delay = config["detrigger_delay"] if "detrigger_delay" in config else DEFAULT_DETRIGGER_DELAY
        self.enable_redis_discovery = config["redis"] if "redis" in config else False
        self.redis_port = config["redis_port"] if "redis_port" in config else DEFAULT_REDIS_PORT
        if "redis_addr" in config:
            self.redis_addr = config["redis_addr"]
        else:
            print("Warning: address of redis server not specified. Defaulting to 'redis_streaming'")
            self.redis_addr = DEFAULT_REDIS_ADDR
        self.is_listening = False
        self.is_triggered = False
        self.redis = None
        self.redis_announcer = None
        self.server_socket = None

    def pre_routine(self):
        # Initialize server socket
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.listen_addr, self.listen_port))
        self.server_socket.listen()
        self.is_listening = True

        # Register at redis
        if self.enable_redis_discovery:
            self.redis_announcer = Thread(target=self.announce_to_redis)
            self.redis_announcer.start()

    def announce_to_redis(self):
        if not NODE_ADDR:
            print("warning: NODE_ADDR not set, streamer won't be reachable from outside")

        self.redis = redis.Redis(self.redis_addr, self.redis_port)
        while self.is_listening:
            try:
                self.redis.set("recorder.%d" % (self.trigger_id),
                            "recorder.%d:%d" % (self.trigger_id, self.listen_port))
                if NODE_ADDR:
                    self.redis.set("recorder.node.%d" % (self.trigger_id),
                                "%s:%d" % (NODE_ADDR, self.listen_port))
            except Exception as e:
                print("Exception occurred setting address on redis:", e)
            finally:
                sleep(10)

    def _trigger(self):
        self.trigger_all()
        sleep(self.detrigger_delay)
        self.detrigger_all()
        self.is_triggered = False

    def trigger(self):
        Thread(target=self._trigger, daemon=True).start()
        
    def routine(self):
        try:
            conn, _ = self.server_socket.accept()
            if not self.is_triggered:
                self.is_triggered = True
                self.trigger()
            conn.close()
        except Exception as e:
            print("Error in network trigger routine:", e)

    def _cleanup(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.connect((self.listen_addr, self.listen_port))
            except socket.error:
                pass

        self.server_socket.close()
        self.is_listening = False
        if self.redis_announcer:
            self.redis_announcer.join()
            self.redis_announcer = None




from time import time
from threading import Thread, Event
from trigger_responses.trigger_response_store import TriggerResponseStore
from utils.consts import GLOBALS, TRIGGER_COOLDOWN_TIME


class BaseTriggerResponse:
    def __init__(self, use_detrigger=False, cooldown_period=GLOBALS[TRIGGER_COOLDOWN_TIME]):
        self.trigger_name = "base"
        self.trigger_count = 0
        self.trigger_event = Event()
        self.detrigger_event = Event()
        self.use_detrigger = use_detrigger
        self.cooldown_period = cooldown_period
        self.last_trigger_time = 0
        self.listening = False
        self.trigger_listener = None
        self.is_running = False
        self.recent_triggers = set()
        self.trigger_args = {}
        self.ready = True

    def start_listening_for_trigger(self):
        """launches thread to listen for trigger.
        """
        if not self.ready:
            print("Trigger Response '%s' is not ready. Not starting." %
                  self.trigger_name)
            return

        self._initialize_trigger_response()
        self.listening = True
        self.trigger_listener = Thread(target=self._listen_for_trigger)
        self.trigger_listener.start()

    def stop_listening_for_trigger(self, wait=True):
        """stops listening to trigger.

        Keyword Arguments:
            wait {bool} -- wether to wait for trigger 
            listening thread to quit this method (default: {True})
        """

        self.listening = False
        self.trigger_event.set()
        if self.trigger_listener and wait:
            self.trigger_listener.join()

        self._cleanup_trigger_response()

    def detrigger(self):
        if self.use_detrigger and self.is_running:
            self.trigger_event.clear()
            self.detrigger_event.set()
            self.is_running = False

    def trigger(self, trigger_name, trigger_args={}):
        if not self.ready:
            print("Trigger Response '%s' is not ready. Not triggering." %
                  self.trigger_name)
            return

        self.recent_triggers.add(trigger_name)
        self.trigger_args.update(trigger_args)

        if not self.is_running:
            self.detrigger_event.clear()
            self.trigger_event.set()
            self.is_running = True

    def _initialize_trigger_response(self):
        pass

    def _cleanup_trigger_response(self):
        pass

    def _trigger_response(self, trigger_args=None):
        """
        Needs to be overriden by actual trigger response
        """
        pass

    def _listen_for_trigger(self):
        while self.listening and self.trigger_event.wait():
            self.trigger_event.clear()
            # don't spam the trigger
            if not time() - self.last_trigger_time < self.cooldown_period:
                self.last_trigger_time = time()
                try:
                    self._trigger_response(self.trigger_args)
                except Exception as e:
                    print("error on trigger response %s:" %
                          self.trigger_name, e)
            self.trigger_args = {}
            self.recent_triggers = set()
            self.is_running = False



import io
from time import sleep
from datetime import datetime
from streamers.streamer import ROOT_PATH
from utils.single_picamera import SinglePiCamera
from .base import BaseTriggerResponse


class Dummy(BaseTriggerResponse):
    def __init__(self, use_detrigger=False, **recording_options):
        super().__init__(use_detrigger=use_detrigger)

    def _trigger_response(self, trigger_args={}):
        print("dummy trigger reponse was triggered with:\n", trigger_args)



class ClientSocketWrap:
    def __init__(self, client_socket, client_address, cleanup=None):
        self.cleanup = cleanup
        self.client_address = client_address
        self.client_socket = client_socket
        self.is_connected = True

    def close(self):
        if self.is_connected:
            self.client_socket.close()
            self.is_connected = False

        if self.cleanup is not None:
            self.cleanup(self.client_address)

    def write(self, data, timeout=5):
        try:
            if self.is_connected:
                self.client_socket.settimeout(timeout)
                self.client_socket.sendall(data)
        except Exception as e:
            print("Exception occurred when writing to client socket:", e)
            self.close()
            return False
        return True



"""
BaseModeExecutor Class
"""

import json
from croniter import croniter
from datetime import datetime


class ModeExecutor:
    def __init__(self, modename, mode):
        self.modename = modename
        self.mode = mode

    def start(self, finish_time=None):
        if finish_time is not None:
            finish_time = croniter(finish_time, datetime.now()).get_next(datetime)

        if not self.mode.is_running:
            if self.mode.initialize():
                self.mode.start(finish_time)
            else:
                raise Exception("Unable to initialize %s mode" % self.modename)
        elif finish_time:
            self.mode.update_finish_time(finish_time)

    def stop(self, wait=True):
        self.mode.shutdown(wait)



import sys
import json
import signal
from streamers.streamer import Streamer
from croniter import croniter
from datetime import datetime
from utils.consts import GLOBALS
from utils.single_picamera import SinglePiCamera
from mode_executor.mode_executor import ModeExecutor
from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.schedulers.blocking import BlockingScheduler
# Import Modes
from modes.motion_detection import MotionDetectionMode
from modes.network_streaming import NetworkStreamingMode
from modes.network_serving import NetworkServingMode
from modes.network_trigger import NetowrkTriggerMode
from modes.photo_motion_detection import PhotoMotionDetection
# Import Trigger Responses
from trigger_responses.dummy import Dummy
from trigger_responses.snapshot import Snapshot
from trigger_responses.record_to_file import RecordToFile
from trigger_responses.http import HTTP
from trigger_responses.trigger_response_store import TriggerResponseStore

SCHEDULES_PATH = "schedules/"
RESOLUTION = "1280x720"
FRAMERATE = 25

MODE_MAP = {
    "motion_detection": MotionDetectionMode,
    "network_streaming": NetworkStreamingMode,
    "network_serving": NetworkServingMode,
    "network_trigger": NetowrkTriggerMode,
    "photo_motion_detection": PhotoMotionDetection
}

TRIGGER_RESPONSE_MAP = {
    "record_to_file": RecordToFile,
    "snapshot": Snapshot,
    "dummy": Dummy
}


def clean_stop(signum, frame):
    # Dummy way to signal the streaming_server configurator to stop
    try:
        with open("stop_configurator", "w") as stop_configurator:
            pass
        scheduler.shutdown(wait=False)
        for executor in mode_executors.values():
            executor[0].stop()
        Streamer.shutdown_streamers()
    except Exception as e:
        print("error exiting:", e)
    finally:
        try:
            SinglePiCamera().close()
        except Exception as e:
            print("failed to close PiCamera cleanly")


def help_message():
    print("./main.py <path_to_config_file>")


def parse_schedule(filename):
    with open(SCHEDULES_PATH + filename + ".cron") as schedule:
        start, finish = list(map(str.strip, schedule.read().split(":")))
    return start, finish


if len(sys.argv) < 2:
    print("Not enough arguments. Need to at least provide the configuration file.")
    help_message()
    sys.exit(1)

with open(sys.argv[1], "r") as config_file:
    config = json.loads(config_file.read())

# Load Globals
if "globals" in config.keys():
    for k, v in config["globals"]:
        GLOBALS[k] = v

if "camera_initialization_options" in config.keys():
    # Initialize singleton with the set configs
    camera = SinglePiCamera(**config["camera_initialization_options"])
else:
    camera = SinglePiCamera(resolution=RESOLUTION, framerate=FRAMERATE)

# Parse and instantiate trigger responses
trigger_response_store = TriggerResponseStore()
if "trigger_responses" in config.keys():
    for trigger_response_name, resp_args in config["trigger_responses"].items():
        if trigger_response_name in TRIGGER_RESPONSE_MAP.keys():
            trigger_response_store.add_trigger_response(trigger_response_name,
                                                        TRIGGER_RESPONSE_MAP[trigger_response_name](**resp_args))

if "modes" not in config.keys():
    print("Config file needs to have 'modes' configuration in order to run.")
    print("Please check the example configurations.")
    sys.exit(1)

modes_configs = config["modes"]

mode_executors = {}
for mode_name, mode_config_list in modes_configs.items():
    for idx, mode_config in enumerate(mode_config_list):
        if "schedules" not in mode_config.keys():
            print("Mode configuration for %s at index %d needs a list of schedules with the 'schedules' key." % (
                mode_name, idx))
            print("Skipping %s..." % mode_name)

        mode_id = "%s_%d" % (mode_name, idx)
        schedules = mode_config["schedules"]
        trigger_responses = mode_config["trigger_responses"] if "trigger_responses" in mode_config else {
        }
        mode_trigger_responses = [trigger_response_store.get_trigger_response(
            name) for name in trigger_responses]
        mode = MODE_MAP[mode_name](
            trigger_responses=mode_trigger_responses, config=mode_config["mode_config"])
        mode_executor = ModeExecutor(mode_id, mode)
        mode_executors[mode_id] = mode_executor, [
            parse_schedule(schedule) for schedule in schedules]

executors = {"default": ThreadPoolExecutor(max_workers=10)}
job_defaults = {"coalesce": True, "max_instances": 1}
scheduler = BlockingScheduler(executors=executors, job_defaults=job_defaults)
for mode, config in mode_executors.items():
    executor, schedules = config
    for sched in schedules:
        # Check if the job is supposed to always run
        # This means it doesn't need to be scheduled
        if sched[0] == sched[1]:
            executor.start()
            continue

        now = datetime.now()
        next_start = croniter(sched[0], now).get_next(datetime)
        next_stop = croniter(sched[1], now).get_next(datetime)
        minute, hour, mday, month, wday = sched[0].split()
        if (next_start - next_stop).total_seconds() >= 0:
            executor.start(sched[1])
            misfire_grace_time = int(
                (next_stop - croniter(sched[0], now).get_prev(datetime)).total_seconds())
        else:
            misfire_grace_time = int((next_stop - next_start).total_seconds())

        if misfire_grace_time == 0:
            misfire_grace_time = None

        # APScheduler and Croniter have different weekday normalization, transforming range into csv of days
        if "-" in wday:
            range_days = list(map(int, wday.split("-")))
            wday = list(map(lambda x: x % 7, list(
                range(range_days[0], range_days[1] + 1))))
            wday = ",".join(map(str, wday))

        # Add job to scheduler
        # The executor.start method checks if the mode is already running
        # If it is it will update its finish_time if it is greater than the current finish_time
        # misfire_grace_time is set to the supposed job duration (stop - start)
        scheduler.add_job(executor.start,
                          args=(sched[1],),
                          trigger="cron",
                          minute=minute,
                          hour=hour,
                          day=mday,
                          month=month,
                          day_of_week=wday,
                          id=mode,
                          misfire_grace_time=misfire_grace_time)

signal.signal(signal.SIGINT, clean_stop)
signal.signal(signal.SIGTERM, clean_stop)
try:
    scheduler.start()
finally:
    SinglePiCamera().close()



import sys
sys.path.append("/picameleon")
from datetime import datetime
from modes.base import BaseMode
from time import sleep, time
import unittest
from unittest.mock import Mock, call

mock_streamer = Mock()
mock_trigger_responses = [Mock() for _ in range(5)]


class TestBaseMode(unittest.TestCase):

    def setUp(self):
        mock_streamer.reset_mock()
        map(lambda x: x.reset_mock, mock_trigger_responses)

    def test_update_finish_time(self):
        """update_finish_time should only update if finish_time is greater than
        """
        mode = BaseMode({}, mock_trigger_responses)
        finish_time = time()
        finish_date = datetime.fromtimestamp(finish_time)
        mode.finish_time = finish_date
        mode.update_finish_time(datetime.fromtimestamp(finish_time + 1))
        self.assertEqual(datetime.fromtimestamp(
            finish_time + 1), mode.finish_time)
        mode.update_finish_time(datetime.fromtimestamp(finish_time - 1))
        self.assertEqual(datetime.fromtimestamp(
            finish_time + 1), mode.finish_time)

    def test_initialize(self):
        mode = BaseMode({}, mock_trigger_responses)
        mode.initialize()
        for mock in mock_trigger_responses:
            mock.assert_has_calls([call.start_listening_for_trigger()])

    def test_start(self):
        mode = BaseMode({}, mock_trigger_responses)
        mode.pre_routine = Mock()
        mode.routine = Mock()
        mode.streamer = mock_streamer
        finish_time = time() + 0.2
        finish_date = datetime.fromtimestamp(finish_time)
        mode.start(finish_date)
        self.assertEqual(finish_date, mode.finish_time)
        self.assertTrue(mode.is_running)
        self.assertIsNotNone(mode.runner_thread)
        self.assertTrue(mode.runner_thread.is_alive())
        mock_streamer.assert_has_calls([call.start()])
        mode.routine.assert_any_call()
        mode.pre_routine.assert_called_once()
        mode.runner_thread.join()

    def test_check_stop(self):
        mode = BaseMode({}, mock_trigger_responses)
        self.assertTrue(mode._check_stop())
        mode.finish_time = datetime.fromtimestamp(time() + .1)
        self.assertTrue(mode._check_stop())
        sleep(.1)
        self.assertFalse(mode._check_stop())

    def test_trigger_all(self):
        mode = BaseMode({}, mock_trigger_responses)
        mode.trigger_all()
        for mock in mock_trigger_responses:
            mock.assert_has_calls([call.trigger("base")])
    """
    def test_retrigger(self):
        mode = BaseMode({}, mock_trigger_responses)
        mock_retrigger = Mock()
        mode.trigger_all(mock_retrigger)
        for mock in mock_trigger_responses:
            mock.assert_has_calls([call.trigger(mock_retrigger)])
        mock_retrigger.assert_called(call._trigger_response())
    """

    def test_detrigger_all(self):
        mode = BaseMode({}, mock_trigger_responses)
        mode.detrigger_all()
        for mock in mock_trigger_responses:
            mock.assert_has_calls([call.detrigger()])

    def test_shutdown(self):
        mode = BaseMode({}, mock_trigger_responses)
        mode._cleanup = Mock()
        mode.runner_thread = Mock()
        mode.streamer = Mock()
        mode.streamer.output = Mock()
        mode.streamer.output.has_outputs = Mock(return_value=False)
        mode.shutdown()
        self.assertFalse(mode.is_running)
        self.assertGreater(datetime.now(), mode.finish_time)
        for mock in mock_trigger_responses:
            mock.assert_has_calls([call.stop_listening_for_trigger(True)])
        mode.runner_thread.assert_has_calls([call.join()])
        mode._cleanup.assert_called_once()
        mode.streamer.output.assert_has_calls([call.has_outputs()])
        mode.streamer.assert_has_calls([call.stop()])


if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append("/picameleon")
from outputs.motion_output_holder import MotionOutputHolder
from outputs.output_holder import OutputHolder, WriterOutputHolder
import time
import unittest
from unittest.mock import Mock, call

mock_output = Mock()


def create_and_start_writer_output_holder():
    output_holder = WriterOutputHolder()
    output_holder.start()
    return output_holder


def write_and_wait(writer_output_holder, sleep, *test_data, analyze=False):
    action = writer_output_holder.analyze if analyze else writer_output_holder.write
    for data in test_data:
        action(data)
    time.sleep(sleep)


class TestOutputHolder(unittest.TestCase):

    def test_add_output(self):
        output_holder = OutputHolder()
        output_holder.add_output("test", mock_output)
        self.assertEqual(1, len(output_holder.outputs))
        self.assertIn("test", output_holder.outputs.keys())

    def test_remove_output(self):
        output_holder = OutputHolder()
        output_holder.add_output("test", mock_output)
        self.assertEqual(1, len(output_holder.outputs))
        output_holder.remove_output("test")
        self.assertEqual(0, len(output_holder.outputs))
        self.assertRaises(KeyError, output_holder.remove_output, "test")

    def test_has_output(self):
        output_holder = OutputHolder()
        output_holder.add_output("test", mock_output)
        self.assertTrue(output_holder.has_output("test"))

    def test_get_output(self):
        output_holder = OutputHolder()
        output_holder.add_output("test", mock_output)
        output = output_holder.get_output("test")
        self.assertIs(output, mock_output)


class TestWriterOutputHolder(unittest.TestCase):

    def setUp(self):
        mock_output.reset_mock()

    def test_write(self):
        """Checks that calls to write on the output 
        handler are correctly passed to the internal outputs
        """

        test_data = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        output_holder = create_and_start_writer_output_holder()
        output_holder.add_output("test1", mock_output)
        write_and_wait(output_holder, .2, *test_data)
        self.assertEqual(mock_output.write.call_count, len(test_data))
        mock_output.assert_has_calls([call.write(data) for data in test_data])
        output_holder.stop()
        output_holder.join()

    def test_stop(self):
        """Checks that no more write calls to the 
        internal outputs are made after closing the output holder
        """

        output_holder = create_and_start_writer_output_holder()
        output_holder.add_output("test", mock_output)
        write_and_wait(output_holder, .1, "test1")
        self.assertEqual(mock_output.write.call_count, 1)
        output_holder.stop()
        time.sleep(.1)
        self.assertFalse(output_holder.is_alive())
        write_and_wait(output_holder, .1, "test1")
        self.assertEqual(mock_output.write.call_count, 1)
        self.assertEqual(len(output_holder.aux_buffer), 0)


class TestMotionOutputHolder(unittest.TestCase):

    def setUp(self):
        mock_output.reset_mock()

    def test_analyze(self):
        test_data = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        output = MotionOutputHolder(outputs={"test": mock_output})
        write_and_wait(output, .1, *test_data, analyze=True)
        self.assertEqual(mock_output.analyze.call_count, len(test_data))
        mock_output.assert_has_calls(
            [call.analyze(data) for data in test_data])


if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append("/picameleon")
from trigger_responses.trigger_response_store import TriggerResponseStore
import unittest
from unittest.mock import Mock, ANY, call, patch, mock_open


class TestHTTP(unittest.TestCase):
    trigger_response = None

    def import_and_initialize(self):
        from trigger_responses.http import HTTP
        self.trigger_response = HTTP(
            http_service_urls={
                "http://test_with_pic.com": True,
                "http://test_without_pic.com": False,
            })

    @patch("requests.post")
    @patch("builtins.open")
    def test_open_and_send(self, mocked_open, mocked_post):
        self.import_and_initialize()
        self.trigger_response._open_and_send(
            "http://test.com/v1/faces?id=the_id.jpg", "/tmp/image/path.jpg")
        mocked_open.assert_called_once_with("/tmp/image/path.jpg", "rb")
        mocked_post.assert_called_once_with("http://test.com/v1/faces?id=the_id.jpg",
                                            files={"file": mocked_open.return_value})
        mocked_open.return_value.assert_has_calls([call.close()])

    @patch("requests.post")
    @patch("builtins.open", new_callable=mock_open, read_data=b'test content')
    def test_trigger_response(self, mocked_open, mocked_post):
        self.import_and_initialize()
        # given
        post_response = Mock()
        post_response.status_code = 200
        mocked_post.return_value = post_response

        # when
        self.trigger_response._trigger_response()

        # then
        self.assertEqual(mocked_post.call_count, 2)
        mocked_post.assert_has_calls([call("http://test_with_pic.com", files=ANY),
                                      call("http://test_without_pic.com")],
                                      any_order=True)

if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append("/picameleon")
from outputs.motion_detector import MotionDetector
from threading import Thread
import numpy
import unittest

THRESHOLD = 8  # Magnitude of motion vector
SENSITIVITY = 10  # Number of motion vectors
DEFAULT_TIMEOUT = 1
motion_detector = MotionDetector(THRESHOLD, SENSITIVITY)
motion_vectors = numpy.load("tests/test_data/motion_vectors.npy")
no_motion_vectors = numpy.load("tests/test_data/no_motion_vectors.npy")


class TestMotionDetectorOutput(unittest.TestCase):

    def setUp(self):
        motion_detector.motion_detected.clear()

    def asynchronous_motion_detection(self, motion_vectors):
        def motion_vectors_through_analyze(motion_vectors):
            for motion_vector in motion_vectors:
                motion_detector.analyze(motion_vector)
        t = Thread(target=motion_vectors_through_analyze,
                   args=(motion_vectors,))
        t.start()
        return t

    def test_analyze_detect_motion(self):
        """Checks if the analyze method detects motion correctly
        """
        job = self.asynchronous_motion_detection(motion_vectors)
        self.assertTrue(motion_detector.wait_for_motion(DEFAULT_TIMEOUT))
        job.join()

    def test_analyze_not_detect_motion(self):
        """Checks if no motion is detected
        """
        job = self.asynchronous_motion_detection(no_motion_vectors)
        self.assertFalse(motion_detector.wait_for_motion(DEFAULT_TIMEOUT))
        job.join()

    def test_calibration_fails(self):
        """Checks if calibration fails when motion is detected
        """
        job = self.asynchronous_motion_detection(motion_vectors)
        self.assertFalse(motion_detector.calibrate(
            no_motion_seconds=DEFAULT_TIMEOUT, max_resets=1))
        job.join()

    def test_calibration_succeeds(self):
        """Checks if calibration fails when motion is detected
        """
        job = self.asynchronous_motion_detection(no_motion_vectors)
        self.assertTrue(motion_detector.calibrate(
            no_motion_seconds=DEFAULT_TIMEOUT, max_resets=1))
        job.join()


if __name__ == '__main__':
    unittest.main()



"""
Base Mode Class.
"""
from threading import Thread
from datetime import datetime


class BaseMode:
    def __init__(self, config, trigger_responses=None):
        self.name = "base"
        self.config = config
        self.finish_time = None
        self.trigger_responses = [] if trigger_responses is None else trigger_responses
        self.is_running = False
        self.runner_thread = None
        self.streamer = None  # Subclass must have a Streamer object here
        self.ready = True

    def update_finish_time(self, finish_time):
        if self.finish_time < finish_time:
            self.finish_time = finish_time

    def initialize(self):
        if not self.ready:
            print("Mode '%s' is not ready. Not initializing." % self.name)
            return

        for trigger_response in self.trigger_responses:
            trigger_response.start_listening_for_trigger()
        return True

    def shutdown(self, wait=True):
        self.is_running = False
        self.finish_time = datetime.now()
        for trigger_response in self.trigger_responses:
            trigger_response.stop_listening_for_trigger(wait)

        self._cleanup()
        if self.runner_thread:
            self.runner_thread.join()

        if self.streamer and not self.streamer.output.has_outputs():
            self.streamer.stop()

    def trigger_all(self):
        for trigger_response in self.trigger_responses:
            trigger_response.trigger(self.name)

    def detrigger_all(self):
        for trigger_response in self.trigger_responses:
            trigger_response.detrigger()

    def start(self, finish_time=None):
        self.finish_time = finish_time
        self.is_running = True
        self.runner_thread = Thread(target=self.run)
        self.runner_thread.start()

    def _check_stop(self):
        if self.finish_time is None:
            return True

        return self.finish_time > datetime.now()

    def run(self):
        if not self.ready:
            print("Mode '%s' is not ready. Not running." % self.name)
            return

        if self.streamer:
            self.streamer.start()

        self.pre_routine()
        while self._check_stop():
            try:
                self.routine()
            except Exception as e:
                print("Caught exception in routine of %s:\n%s" %
                      (type(self).__name__, str(e)))

    def pre_routine(self):
        pass

    def routine(self):
        """
        Write code run in routine
        """
        pass

    def _cleanup(self):
        pass



import sys
sys.path.append("/picameleon")
from unittest.mock import patch
import unittest
import filecmp
import time
import os
from streamers.base import BaseStreamer
from streamers.split_frame import SplitFrameStreamer
from streamers.streamer import Streamer


class TestStreamer(unittest.TestCase):

    def test_multiple_output_write(self):
        """Tests if writing to multiple outputs produces 3 equally sized files
        """

        output1 = open("out1.h264", "wb")
        output2 = open("out2.h264", "wb")
        output3 = open("out3.h264", "wb")

        streamer1 = Streamer("h264", id_output=("1", output1))
        streamer2 = Streamer("h264", id_output=("2", output2))
        streamer3 = Streamer("h264", id_output=("3", output3))

        self.assertIs(streamer1, streamer2)
        self.assertIs(streamer2, streamer3)
        self.assertEqual(len(streamer1.output.outputs), 3)
        self.assertEqual(len(Streamer.get_available_ports()), 3)

        streamer1.start()
        streamer1.wait_recording(1)
        streamer1.stop()
        time.sleep(0.5)

        output1.close()
        output2.close()
        output3.close()

        self.assertEqual(os.stat("out1.h264").st_size,
                         os.stat("out2.h264").st_size)
        self.assertEqual(os.stat("out2.h264").st_size,
                         os.stat("out3.h264").st_size)
        self.assertEqual(filecmp.cmp("out1.h264", "out2.h264"), True)
        self.assertEqual(filecmp.cmp("out2.h264", "out3.h264"), True)

        os.remove("out1.h264")
        os.remove("out2.h264")
        os.remove("out3.h264")

    def test_multi_streamer_singleton(self):
        streamer1 = Streamer("h264")
        self.assertEqual(len(Streamer.get_available_ports()), 3)
        streamer2 = Streamer("mjpeg")
        self.assertEqual(len(Streamer.get_available_ports()), 2)
        streamer3 = Streamer("yuv420")
        self.assertEqual(len(Streamer.get_available_ports()), 1)
        streamer4 = Streamer("h264", recording_options={"resize": "1280x720"})
        self.assertEqual(len(Streamer.get_available_ports()), 0)

        self.assertIsNot(streamer1, streamer2)
        self.assertIsNot(streamer1, streamer3)
        self.assertIsNot(streamer1, streamer4)
        self.assertIsNot(streamer2, streamer3)
        self.assertIsNot(streamer2, streamer4)
        self.assertIsNot(streamer3, streamer4)

        self.assertIsInstance(streamer1, SplitFrameStreamer)
        self.assertIsInstance(streamer2, SplitFrameStreamer)
        self.assertIsInstance(streamer3, BaseStreamer)
        self.assertIsInstance(streamer4, SplitFrameStreamer)
        self.assertRaises(Exception, Streamer, "mjpeg", recording_options={"resize": "1920x1080"})

        streamer4.stop()
        self.assertEqual(len(Streamer.get_available_ports()), 1)
        streamer4 = Streamer("mjpeg", recording_options={"resize": "1920x1080"})
        self.assertEqual(len(Streamer.get_available_ports()), 0)
        self.assertRaises(Exception, Streamer, "mjpeg", recording_options={"resize": "480p"})
        streamer1.stop()
        self.assertEqual(len(Streamer.get_available_ports()), 1)
        streamer2.stop()
        self.assertEqual(len(Streamer.get_available_ports()), 2)
        streamer3.stop()
        self.assertEqual(len(Streamer.get_available_ports()), 3)
        streamer4.stop()
        self.assertEqual(len(Streamer.get_available_ports()), 4)

    def test_get_picture_port(self):
        # Check initial port availability
        self.assertEqual(len(Streamer.get_available_ports()), 4)

        # Get a picture port and check port availability
        picture_port = Streamer._get_picture_port()
        self.assertEqual(len(Streamer.get_available_ports()), 3)

        # Get another picture port and check port availability and that it's same port
        repicture_port = Streamer._get_picture_port()
        self.assertEqual(len(Streamer.get_available_ports()), 3)
        self.assertEqual(picture_port, repicture_port)

        # Return the picture port and check that availability has been restored
        Streamer._return_port_availability(picture_port)
        self.assertIsNone(Streamer._picture_port)
        self.assertEqual(len(Streamer.get_available_ports()), 4)

    @patch("picamerax.PiCamera.capture")
    def test_take_picture(self, mocked_capture):
        # Check initial port availability and port being set
        self.assertEqual(len(Streamer.get_available_ports()), 4)
        self.assertIsNone(Streamer._picture_port)

        # Take picture with all possible extra args
        Streamer.take_picture("test_output.jpg",
                              format="randomFormat",
                              **{"option1": 1, "option2": 2, "resize": (100,200)})

        # Check everything was returned after picture was taken
        self.assertEqual(len(Streamer.get_available_ports()), 4)
        self.assertIsNone(Streamer._picture_port)

        # Check picamerax.capture was called correctly
        mocked_capture.assert_called_once_with('test_output.jpg',
                                               format='randomFormat',
                                               option1=1,
                                               option2=2,
                                               resize=(100, 200),
                                               splitter_port=2,
                                               use_video_port=True)


if __name__ == '__main__':
    unittest.main()



"""
SinglePiCamera class

Since only one instance of PiCamera can exist this
simply wraps the PiCamera object in a singleton class.
"""
import inspect

from picamerax import PiCamera


class SinglePiCamera:
    instance = None

    def __init__(self, **options):
        if not SinglePiCamera.instance:
            picam_args = set(inspect.getfullargspec(PiCamera).args)

            # clean up args that dont belong and set new ones that didn't exist before
            args = {k: v for k, v in options.items() if k in picam_args}
            options = {k: v for k, v in options.items() if k not in picam_args}

            SinglePiCamera.instance = PiCamera(**args)
            for k, v in options.items():
                setattr(SinglePiCamera.instance, k, v)

    def __getattr__(self, attr):
        return getattr(SinglePiCamera.instance, attr)

    def __setattr__(self, name, value):
        return setattr(SinglePiCamera.instance, name, value)



import socket
import struct
from threading import Lock

SOCKET_TYPES = {
    "udp": socket.SOCK_DGRAM,
    "tcp": socket.SOCK_STREAM
}


class SocketWrap:
    def __init__(self, host, port, socket_type, greeting=None):
        self.socket = socket.socket(socket.AF_INET, SOCKET_TYPES[socket_type])
        self.sock_file = None
        self._is_connected = False
        self.socket_type = socket_type
        self.host = host
        self.port = port
        self.greeting = greeting
        self._lock = Lock()

    def is_connected(self):
        if self.socket_type == "tcp":
            return self._is_connected
        return True

    def _send_greeting(self):
        for data in self.greeting:
            if type(data) is int:
                self.sock_file.write(struct.pack('<L', data))
            elif type(data) is str:
                self.sock_file.write(data.encode("utf-8"))
            elif type(data) is bytes:
                self.sock_file.write(data)
            else:
                print("Don't know how to send data with type:", type(data))
        self.sock_file.flush()

    def connect(self):
        with self._lock:
            if not self.is_connected():
                self.socket = socket.socket(socket.AF_INET, SOCKET_TYPES[self.socket_type])
                self.socket.settimeout(5)
                try:
                    self.socket.connect((self.host, self.port))
                    self.sock_file = self.socket.makefile("wb")
                    if self.greeting:
                        self._send_greeting()
                    self._is_connected = True
                    return True
                except Exception as e:
                    print("Error on connection:", e)
                    self._is_connected = False
                    return False
            return True

    def close(self):
        if self.is_connected() and self.socket_type == "tcp":
            self.sock_file.close()
            self.socket.close()

    def write(self, data):
        with self._lock:
            try:
                if self.is_connected():
                    if self.socket_type == "tcp":
                        self.sock_file.write(data)
                        self.sock_file.flush()
                    else:
                        while len(data) > 1024:
                            to_send = data[:1024]
                            self.socket.sendto(to_send, (self.host, self.port))
                            data = data[1024:]
                        self.socket.sendto(data, (self.host, self.port))
            except Exception as e:
                print("Error occured, socket disconnected.", e)
                self._is_connected = False
                return False
        return True



"""
The Photo Motion Detection mode.

It just quickly takes pictures and compares them with the previous ones

When motion is detected it will use the configured trigger response.
"""

import picamerax.array
from .base import BaseMode
from utils.single_picamera import SinglePiCamera
from streamers.streamer import Streamer

WIDTH = 224  # Default Width
HEIGHT = 160  # Default Height
# How Much pixel changes (really depends a lot on exposure_mode and awb_mode, this value seems to be good for default)
THRESHOLD = 100
SENSITIVITY = 50  # How many pixels change
MIN_CERTAINTY = 3  # How often the conditions must apply in a row
MOTION_DETECTOR_DETRIGGER = 5  # Number of pics without motion to detrigger


class PhotoMotionDetection(BaseMode):
    def __init__(self, config: dict, trigger_responses):
        super().__init__(config, trigger_responses)
        self.name = "photo_motion_detection"

        # Motion Detection Parameters
        if "resize" not in config.keys():
            config["resize"] = (WIDTH, HEIGHT)

        self.threshold = config["threshold"] if "threshold" in config else THRESHOLD
        self.sensitivity = config["sensitivity"] if "sensitivity" in config else SENSITIVITY
        self.min_certainty = config["min_certainty"] if "min_certainty" in config else MIN_CERTAINTY
        self.pic_buffer = picamerax.array.PiRGBArray(SinglePiCamera(), size=(config["resize"][0], config["resize"][1]))
        self.prev_pic = None
        self.current_certainty = 0
        self.motion_detected = False
        self.iterations_without_motion = 0

    def get_rgb_pic(self):
        self.pic_buffer.truncate(0)
        Streamer.take_picture(self.pic_buffer, format='rgb', capture_options=self.config)
        return self.pic_buffer.array

    def pre_routine(self):
        # Get first pic to compare to
        self.prev_pic = self.get_rgb_pic()

    def routine(self):
        if self._detect_motion():
            self.motion_detected = True
            self.trigger_all()
            self.iterations_without_motion = 0
        elif self.motion_detected:
            self.iterations_without_motion += 1
            if self.iterations_without_motion >= MOTION_DETECTOR_DETRIGGER:
                self.iterations_without_motion = 0
                self.detrigger_all()
                self.motion_detected = False

    def _detect_motion(self):
        # Take new pic
        new_pic = self.get_rgb_pic()

        # Simple Motion Detection Algorithm
        diff_count = 0
        found_motion_this_pic = False
        for w in range(0, self.config["resize"][0]):
            if found_motion_this_pic:
                break
            for h in range(0, self.config["resize"][1]):
                # get the diff of the pixel. Conversion to int
                # is required to avoid unsigned short overflow.
                diff_r = abs(int(self.prev_pic[h][w][0]) - int(new_pic[h][w][0]))
                diff_g = abs(int(self.prev_pic[h][w][1]) - int(new_pic[h][w][1]))
                diff_b = abs(int(self.prev_pic[h][w][2]) - int(new_pic[h][w][2]))
                total_diff = diff_r + diff_g + diff_b
                if total_diff > self.threshold:
                    diff_count += 1
                    if diff_count > self.sensitivity:
                        self.current_certainty += 1
                        found_motion_this_pic = True
                        break

        if self.current_certainty > self.min_certainty:
            self.current_certainty = 0
            return True

        if diff_count < self.sensitivity and self.current_certainty > 0:
            self.current_certainty -= 1
        self.prev_pic = new_pic
        return False

    def _cleanup(self):
        self.detrigger_all()



class TriggerResponseStore:
    __instance = None
    trigger_responses = None

    def __new__(cls):
        if TriggerResponseStore.__instance is None:
            TriggerResponseStore.__instance = object.__new__(cls)
            TriggerResponseStore.__instance.trigger_responses = {}
        return TriggerResponseStore.__instance

    def add_trigger_response(self, name, trigger_response):
        self.trigger_responses[name] = trigger_response

    def get_trigger_response(self, name):
        return self.trigger_responses[name] \
            if name in self.trigger_responses.keys() else None



"""
This is the OutputHolder class.

This class helps streamers reuse camera ports for the same format and size combination
by passing the captured content through the various output analyzers
"""

from .output_holder import OutputHolder
from utils.single_picamera import SinglePiCamera
from picamerax.array import PiMotionAnalysis


class MotionOutputHolder(OutputHolder, PiMotionAnalysis):
    def __init__(self, outputs={}, size=None):
        OutputHolder.__init__(self, outputs)
        PiMotionAnalysis.__init__(self, SinglePiCamera(), size=size)

    def analyze(self, motion_vectors):
        futures = []
        with self.output_lock:
            for _, output in self.outputs.items():
                futures.append(self.pool.submit(output.analyze, motion_vectors))
            for future in futures:
                future.result()


import io
from time import sleep
from datetime import datetime
from streamers.streamer import ROOT_PATH, Streamer
from utils.single_picamera import SinglePiCamera
from .base import BaseTriggerResponse


class Snapshot(BaseTriggerResponse):
    def __init__(self, port, format, destination_path,
                 use_detrigger=False, **capture_options):
        super().__init__(use_detrigger=use_detrigger)
        self.format = format
        self.capture_options = capture_options
        self.destination_path = destination_path if destination_path.startswith(
            "/") else ROOT_PATH + destination_path
        self.camera = SinglePiCamera()

    def _trigger_response(self, trigger_args={}):
        now = datetime.now()
        Streamer.take_picture(self.destination_path + now.strftime("%Y-%m-%d_%H-%M-%S") + "." + self.format,
                              format=self.format, **self.capture_options)



import sys
sys.path.append("/picameleon")
from unittest.mock import Mock
from modes.photo_motion_detection import PhotoMotionDetection
import numpy
import unittest

WIDTH = 224  # Default Width
HEIGHT = 160  # Default Height
# How Much pixel changes (really depends a lot on exposure_mode and awb_mode, this value seems to be good for default)
THRESHOLD = 100
SENSITIVITY = 50  # How many pixels change
MIN_CERTAINTY = 3  # How often the conditions must apply in a row
MOTION_DETECTOR_DETRIGGER = 5  # Number of pics without motion to detrigger
photo_motion_detector = None
base_pic = numpy.load("tests/test_data/base_pic.rgb.npy")
motion_pic = numpy.load("tests/test_data/motion_pic.rgb.npy")


def get_always_same_pic():
    return base_pic


get_base = False


def get_alternate_pics():
    global get_base
    ret = base_pic if get_base else motion_pic
    get_base = not get_base
    return ret


class TestMotionDetectorOutput(unittest.TestCase):

    def setUp(self):
        global photo_motion_detector
        photo_motion_detector = PhotoMotionDetection({}, {})
        photo_motion_detector.detrigger_all = Mock()
        photo_motion_detector.iterations_without_motion = 0
        photo_motion_detector.current_iteration = 0
        photo_motion_detector.current_certainty = 0
        photo_motion_detector.min_certainty = MIN_CERTAINTY
        photo_motion_detector.motion_detected = False
        photo_motion_detector.prev_pic = base_pic

    def test_not_detect_motion(self):
        """Checks if the motion detection algorithm works by
           not finding motion when always passing same config
        """
        photo_motion_detector.get_rgb_pic = get_always_same_pic
        for _ in range(10):
            photo_motion_detector.routine()
            self.assertEqual(photo_motion_detector.current_certainty, 0)
            self.assertFalse(photo_motion_detector.motion_detected)

    def test_detect_motion_and_detrigger(self):
        """Checks if the motion detection algorithm works by
           not finding motion when always passing same config
        """
        photo_motion_detector.get_rgb_pic = get_alternate_pics
        for i in range(MIN_CERTAINTY):
            photo_motion_detector.routine()
            self.assertEqual(photo_motion_detector.current_certainty, i + 1)
            self.assertFalse(photo_motion_detector.motion_detected)

        # Run routine 1 more time to detect motion
        photo_motion_detector.routine()
        self.assertEqual(photo_motion_detector.current_certainty, 0)
        self.assertTrue(photo_motion_detector.motion_detected)

        # Now don't detect until detriggered
        photo_motion_detector.prev_pic = base_pic
        photo_motion_detector.get_rgb_pic = get_always_same_pic
        for i in range(1, MOTION_DETECTOR_DETRIGGER):
            photo_motion_detector.routine()
            self.assertEqual(
                photo_motion_detector.iterations_without_motion, i)
            self.assertTrue(photo_motion_detector.motion_detected)

        # Run routine 1 more time to detrigger motion detection
        photo_motion_detector.routine()
        self.assertEqual(photo_motion_detector.iterations_without_motion, 0)
        self.assertFalse(photo_motion_detector.motion_detected)
        self.assertTrue(photo_motion_detector.detrigger_all.called)


if __name__ == '__main__':
    unittest.main()



"""
The Motion Detection mode.

It uses the SplitFrameStreamer with the mjpeg format to
continuously capture and compare images.

When motion is detected it will use the configured trigger response.
"""

from .base import BaseMode
from outputs.motion_detector import MotionDetector, OUTPUT_ID
from utils.single_picamera import SinglePiCamera
from streamers.streamer import Streamer


class MotionDetectionMode(BaseMode):
    def __init__(self, config, trigger_responses):
        super().__init__(config, trigger_responses)
        self.name = "motion_detection"
        self.recording_options = config["recording_options"] if "recording_options" in config else {}

        self.streamer = Streamer("h264", recording_options=self.recording_options)
        # TODO: Add Motion Detector Configs
        if self.streamer.motion_output.has_output(OUTPUT_ID):
            self.motion_detector = self.streamer.motion_output.get_output(OUTPUT_ID)
        else:
            self.motion_detector = MotionDetector()
            self.streamer.motion_output.add_output(OUTPUT_ID, self.motion_detector)

    def pre_routine(self):
        self.motion_detector.calibrate()

    def routine(self):
        # Wait until motion is detected
        if self.motion_detector.wait_for_motion(5):
            self.trigger_all()
            # Wait until motion is no longer detected
            while not self.motion_detector.calibrate(no_motion_seconds=5):
                pass
            self.detrigger_all()

    def _cleanup(self):
        self.streamer.motion_output.remove_output(OUTPUT_ID)



import os
import json
import redis
import socket
import struct
from time import sleep
from threading import Thread, Lock
from .base import BaseMode
from streamers.streamer import Streamer
from outputs.client_socket_wrap import ClientSocketWrap
from queue import Queue

DEFAULT_MAX_STREAMS = 2
DEFAULT_REDIS_ADDR = "redis_streaming"
DEFAULT_REDIS_PORT = 6379
DEFAULT_LISTEN_ADDR = "0.0.0.0"
DEFAULT_LISTEN_PORT = 5555
STREAM_ID = int(os.getenv("STREAMER_ID", 0))
NODE_ADDR = os.getenv("NODE_ADDR", "")


class NetworkServingMode(BaseMode):
    def __init__(self, config, trigger_responses):
        super().__init__(config, trigger_responses)
        self.name = "network_serving"
        self.stream_id = config["stream_id"] if "stream_id" in config else STREAM_ID
        self.listen_addr = config["listen_addr"] if "listen_addr" in config else DEFAULT_LISTEN_ADDR
        self.listen_port = config["listen_port"] if "listen_port" in config else DEFAULT_LISTEN_PORT
        self.max_streams = config["max_streams"] if "max_streams" in config else DEFAULT_MAX_STREAMS
        self.enable_redis_discovery = config["redis"] if "redis" in config else False
        self.redis_port = config["redis_port"] if "redis_port" in config else DEFAULT_REDIS_PORT
        if "redis_addr" in config:
            self.redis_addr = config["redis_addr"]
        else:
            print("Warning: address of redis server not specified. Defaulting to 'redis_streaming'")
            self.redis_addr = DEFAULT_REDIS_ADDR
        self.is_listening = False
        self.redis = None
        self.redis_announcer = None
        self.server_socket = None
        self.socket_map = {}
        self.socket_to_stream = {}
        self.streamer_map = {}
        self._lock = Lock()
        self._unregister_queue = Queue()

        # Streamer init options
        self.prepend_size = config["prepend_size"] if "prepend_size" in config else False
        self.recording_options = config["recording_options"] if "recording_options" in config else {}

    def pre_routine(self):
        # Initialize server socket
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.listen_addr, self.listen_port))
        self.server_socket.listen()
        self.is_listening = True
        
        # Register at redis
        if self.enable_redis_discovery:
            self.redis_announcer = Thread(target=self.announce_to_redis)
            self.redis_announcer.start()

    def announce_to_redis(self):
        if not NODE_ADDR:
            print("warning: NODE_ADDR not set, streamer won't be reachable from outside")

        self.redis = redis.Redis(self.redis_addr, self.redis_port)
        while self.is_listening:
            try:
                self.redis.set("streamer.%d" % self.stream_id,
                               "streamer.%d:%d" % (self.stream_id, self.listen_port))
                if NODE_ADDR:
                    self.redis.set("streamer.node.%d" % self.stream_id,
                                   "%s:%d" % (NODE_ADDR, self.listen_port))
            except Exception as e:
                print("Exception occured setting address on redis:", e)
            finally:
                sleep(10)

    def cleanup_output(self, output_id):
        with self._lock:
            del self.socket_map[output_id]
            if output_id in self.socket_to_stream:
                streamer_id = self.socket_to_stream[output_id]
                # at this point the output hasn't been removed yet because the callback didn't return, so the number of outputs should be 1
                if streamer_id in self.streamer_map and len(self.streamer_map[streamer_id].output.outputs) == 1:
                    self.streamer_map[streamer_id].stop()
                    del self.streamer_map[streamer_id]
                del self.socket_to_stream[output_id]

    def parse_request(self, conn: socket) -> dict:
        conn.settimeout(5)
        input_length = struct.unpack('<L', conn.recv(4))[0]
        request = json.loads(conn.recv(input_length))
        return request

    def can_serve_client(self, stream_key: str, options = None) -> bool:
        default_options = {"format": "h264"}
        if options is None:
            options = default_options
        else:
            options = {**default_options, **options}

        if stream_key not in self.streamer_map:
            if len(self.streamer_map) >= self.max_streams:
                return False

            try:
                vformat = options["format"]
                del options["format"]
                self.streamer_map[stream_key] = Streamer(vformat,
                                                         prepend_size=self.prepend_size,
                                                         recording_options={
                                                             **self.recording_options,
                                                             **options
                                                             })
            except Exception as e:
                print("unable to serve client probably because no more video ports are available:", e)
                return False
        return True

    def routine(self):
        conn, client_address = self.server_socket.accept()
        stream_key = ""
        try:
            options = self.parse_request(conn)
            stream_key = streamer_key(options)
            with self._lock:
                if not self.can_serve_client(stream_key, options):
                    conn.sendall(struct.pack('<L', 0))
                    conn.close()
                    return

                client_socket = ClientSocketWrap(conn, client_address, self.cleanup_output)
                self.socket_map[client_address] = client_socket
                self.socket_to_stream[client_address] = stream_key

                streamer = self.streamer_map[stream_key]
                if len(self.socket_map) >= 1 and not streamer.is_running:
                    streamer.start()

                streamer.output.add_output(client_address, client_socket)
        except Exception as e:
            print("Error in network serving routine:", e)
            if client_address in self.socket_map.keys():
                self.socket_map[client_address].close()
            else:
                conn.close()

            if client_address in self.socket_to_stream.keys():
                del self.socket_to_stream[client_address]

            if stream_key in self.streamer_map.keys():
                streamer.stop()
                del self.streamer_map[stream_key]

    def _cleanup(self):
        sockets_to_clean = [s for s in self.socket_map.values()]
        for sock in sockets_to_clean:
            sock.close()

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.connect((self.listen_addr, self.listen_port))
            except:
                pass

        self.socket_map = {}
        self.socket_to_stream = {}
        self.server_socket.close()
        self.is_listening = False
        if self.redis_announcer:
            self.redis_announcer.join()
            self.redis_announcer = None


def streamer_key(options) -> str:
    stream_key = ""
    for key in sorted(options.keys()):
        stream_key += "%s_%s" % (key, str(options[key]))
    return stream_key



"""
BaseStreamer is the base class for generic streamer implementations.
"""
from utils.single_picamera import SinglePiCamera
from outputs.output_holder import WriterOutputHolder
from outputs.motion_output_holder import MotionOutputHolder
from threading import Thread, Lock


class BaseStreamer:
    def __init__(self, port, format, options):
        self.camera = SinglePiCamera()
        self.port = port
        self.format = format
        self.resize = self.parse_resize(options["resize"])
        del options["resize"]
        self.output = WriterOutputHolder()
        self.motion_output = MotionOutputHolder(size=self.resize)
        self.sub_output = None
        self.options = options
        self.is_running = False
        self.is_recording = False
        self._callback = lambda port: None
        self._lock = Lock()

    def parse_resize(self, resize):
        try:
            if type(resize) is str:
                width, height = map(int, resize.split("x"))
                return (width, height)
            if type(resize) is tuple or type(resize) is list:
                return (resize[0], resize[1])
        except Exception as e:
            print("error parsing resize parameter:", e)

        return (self.camera.resolution[0], self.camera.resolution[1])

    def close(self):
        try:
            if self.is_recording:
                print("trying to stop recording on port %d" % self.port)
                self.camera.stop_recording(splitter_port=self.port)
        except Exception as e:
            print("Error stopping recording on port %d: %s" % (self.port, e))
        finally:
            self.is_recording = False
            print("stopped recording on port %d" % self.port)

    def stop(self):
        with self._lock:
            self._callback(self.port)
            if not self.is_running:
                return

            self.close()
            self.is_running = False
            if self.motion_output:
                self.motion_output.shutdown(True)
            if self.output and type(self.output) is WriterOutputHolder:
                self.output.stop()
                self.output.shutdown(True)
            self._teardown_streamer()

    def start(self):
        with self._lock:
            if not self.is_running:
                self.is_running = True
                self._setup_streamer()
                if type(self.output) is WriterOutputHolder and not self.output.writable:
                    self.output.start()
                self.run_streamer()

    def run_streamer(self):
        output = self.sub_output if self.sub_output else self.output
        print("Start recording with %s format, size: %s, on port %d" % (self.format, str(self.resize), self.port))
        if self.format == "h264":
            self.options["motion_output"] = self.motion_output

        if self.resize[0] == self.camera.resolution[0] and self.resize[1] == self.camera.resolution[1]:
            self.resize = None

        if not self.is_recording:
            self.camera.start_recording(output,
                                        format=self.format,
                                        resize = self.resize,
                                        splitter_port=self.port,
                                        **self.options)
            print("started recording on port %d" % self.port)
            self.is_recording = True

    def wait_recording(self, seconds):
        self.camera.wait_recording(seconds, self.port)

    def split_recording(self, output):
        """
        Only use this method if you have changed the default output of OutputHolder
        """
        self.camera.split_recording(output, splitter_port=self.port)
        self.output = output

    def rebind_output(self, output_id, output):
        self.output.prepare_new_output(output_id, output)
        self.camera.split_recording(self.output, splitter_port=self.port)
        self.output.notify_split()

    def _setup_streamer(self):
        pass

    def _teardown_streamer(self):
        pass

    def shutdown(self):
        self.stop()
        if self.motion_output:
            self.motion_output.shutdown()
        if type(self.output) is WriterOutputHolder:
                self.output.stop()
                self.output.shutdown()




import sys
sys.path.append("/picameleon")
from mode_executor.mode_executor import ModeExecutor
from datetime import datetime
import unittest
from unittest.mock import Mock, call

mode = Mock()


def datetime_minutes(now=None):
    if not now:
        now = datetime.now()

    return datetime(now.year,
                    now.month,
                    now.day,
                    now.hour,
                    min(59, now.minute + 1))


class TestModeExecutor(unittest.TestCase):

    def setUp(self):
        mode.reset_mock()

    def test_start_mode_not_running(self):
        """tests the start method when mode is not running
        """
        mode.is_running = False
        mode.initialize = Mock(return_value=True)
        mexec = ModeExecutor("test", mode)
        mexec.start("* * * * *")
        mode.assert_has_calls([call.initialize(),
                               call.start(datetime_minutes())])

    def test_start_mode_is_running(self):
        """tests the start method when mode is running
        """
        mode.is_running = True
        mexec = ModeExecutor("test", mode)
        mexec.start("* * * * *")
        mode.assert_has_calls([call.update_finish_time(datetime_minutes())])

    def test_start_mode_init_error(self):
        """tests the start method when mode fails to initialize
        """
        mode.is_running = False
        mode.initialize = Mock(return_value=False)
        mexec = ModeExecutor("test", mode)
        self.assertRaises(Exception, mexec.start, "* * * * *")

    def test_stop(self):
        """tests stop method
        """
        mexec = ModeExecutor("test", mode)

        # With wait defaulting to True
        mexec.stop()
        mode.assert_has_calls([call.shutdown(True)])

        # With wait set to False
        mexec.stop(False)
        mode.assert_has_calls([call.shutdown(False)])


if __name__ == '__main__':
    unittest.main()



"""
Generic Streamer Class that returns one of the implementations of base_streamer.
It also works as a multi-singleton class as it will only create new Streamer instances if they have a new output format or resolution
Different outputs can be appended to the same streamer instance.

If the formats "h264" or "mjpeg" are set then it will return a SplitFrameStreamer which streams a frame at a time.
All other formats return an instance of BaseStreamer which produces raw streams.
"""
import os
from .base import BaseStreamer
from .split_frame import SplitFrameStreamer
from utils.single_picamera import SinglePiCamera
from threading import Lock

ROOT_PATH = '{0}/../'.format(os.path.dirname(os.path.abspath(__file__)))
SPLIT_FRAME_FORMATS = ("mjpeg", "h264")


def get_streamer_instance(port, format, prepend_size, split_frames, recording_options):
    if split_frames:
        return SplitFrameStreamer(port, format, prepend_size, options=recording_options)
    else:
        return BaseStreamer(port, format, options=recording_options)


def ordered_dict_hash(dictionary):
    ordered_keys = sorted(dictionary.keys())
    to_hash = ""
    for key in ordered_keys:
        to_hash += str(key) + str(dictionary[key])
    return hash(to_hash)


class Streamer:
    _lock = Lock()
    _capture_lock = Lock()
    _available_ports = [0, 1, 2, 3]
    _picture_port = None
    _streamers = {}
    _portmap = {}
    _pic_holder = {}

    def __new__(cls, format,
                id_output=(None, None),
                id_motion_output=(None, None),
                prepend_size=False,
                split_frames=True,
                recording_options=None):
        with cls._lock:
            if recording_options is None:
                recording_options = {"resize": SinglePiCamera().resolution}
            elif "resize" not in recording_options.keys():
                recording_options = {"resize": SinglePiCamera().resolution, **recording_options}

            resolution = recording_options["resize"]
            res_str = "_".join(map(str, resolution))
            port_key = "%s_%s_%s_%s" % (format, res_str, str(split_frames), ordered_dict_hash(recording_options))
            outID, output = id_output[0], id_output[1]
            moutID, motion_output = id_motion_output[0], id_motion_output[1]
            if port_key in cls._streamers.keys():
                streamer = cls._streamers[port_key]
                if output:
                    streamer.output.add_output(outID, output)
                if motion_output:
                    streamer.motion_output.add_output(moutID, motion_output)
                return streamer

            # Try cleaning up streamers and recovering ports
            elif len(cls._available_ports) == 0:
                raise Exception("No more camera ports available for new streamer")

            port = cls._available_ports.pop(0)
            new_streamer = get_streamer_instance(
                port, format, prepend_size, split_frames, recording_options)
            if output and outID:
                new_streamer.output.add_output(outID, output)
            if motion_output and moutID:
                new_streamer.motion_output.add_output(moutID, motion_output)

            cls._streamers[port_key] = new_streamer
            cls._portmap[port] = port_key
            new_streamer._callback = cls._return_port_availability
            return new_streamer

    @classmethod
    def shutdown_streamers(cls):
        keys = [key for key in cls._streamers.keys()]
        for streamerId in keys:
            cls._streamers[streamerId].stop()
            cls._streamers[streamerId].shutdown()


    @classmethod
    def get_available_ports(cls):
        with cls._lock:
            return cls._available_ports

    @classmethod
    def _get_picture_port(cls):
        """Picture ports can be reused between components as they don't need
        constant hold on it as video recorders do.
        """
        with cls._lock:
            if cls._picture_port is not None:
                return cls._picture_port

            if len(cls._available_ports) > 0:
                cls._picture_port = cls._available_ports.pop(0)

            return cls._picture_port

    @classmethod
    def _return_port_availability(cls, port):
        with cls._lock:
            if port in cls._portmap:
                port_key = cls._portmap[port]
                del cls._portmap[port]
                del cls._streamers[port_key]
                cls._available_ports.append(port)
            elif port == cls._picture_port:
                cls._available_ports.append(port)
                cls._picture_port = None

    @classmethod
    def take_picture(cls, output, format="jpeg", **capture_options):
        """Picture ports can be reused between components that have the same
        requirements as they don't need constant hold on it as video recorders do.
        """
        with cls._capture_lock:
            port = cls._get_picture_port()
            if port is not None:
                SinglePiCamera().capture(output,
                                         splitter_port=port,
                                         format=format,
                                         use_video_port=True,
                                         **capture_options)
                cls._return_port_availability(port)



import sys
sys.path.append("/picameleon")
import unittest
from unittest.mock import patch, ANY


class TestSnapshot(unittest.TestCase):

    @patch("picamerax.PiCamera.capture")
    def test_trigger_response(self, mocked_capture):
        from trigger_responses.snapshot import Snapshot
        trigger_reponse = Snapshot(port=0,
                                   format="jpeg",
                                   destination_path="/tmp/")
        trigger_reponse._trigger_response()
        mocked_capture.assert_called_once_with(ANY,
                                               format='jpeg',
                                               splitter_port=0,
                                               use_video_port=True)


if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append("/picameleon")
from streamers.streamer import Streamer, ROOT_PATH
from utils.single_picamera import SinglePiCamera
from streamers.streamer import Streamer
import unittest
from unittest.mock import MagicMock, call, patch, ANY

SECONDS_BEFORE = 5
SECONDS_AFTER = 5


class TestRecordToFile(unittest.TestCase):

    trigger_reponse = None

    def import_and_initialize(self):
        from trigger_responses.record_to_file import RecordToFile
        self.trigger_reponse = RecordToFile(format="h264",
                                            destination_path="captures",
                                            record_time_before_trigger=SECONDS_BEFORE,
                                            record_time_after_trigger=SECONDS_AFTER,
                                            use_detrigger=False)

    @patch("utils.single_picamera.SinglePiCamera")
    @patch("picamerax.PiCameraCircularIO")
    @patch("streamers.streamer.Streamer")
    def test_initialize_trigger_response(self, mocked_streamer, mocked_picam_io, mocked_picam):
        self.import_and_initialize()
        self.trigger_reponse._initialize_trigger_response()
        mocked_streamer.assert_called_once_with("h264",
                                                split_frames=False,
                                                recording_options={})
        mocked_picam_io.assert_called_once_with(
            ANY, seconds=SECONDS_BEFORE, splitter_port=ANY)
        self.assertEqual(self.trigger_reponse.streamer.output,
                         self.trigger_reponse.before_buffer)
        self.trigger_reponse.streamer.assert_has_calls([call.start()])

    def test_trigger_response(self):
        self.import_and_initialize()
        self.trigger_reponse.streamer = MagicMock()
        self.trigger_reponse.before_buffer = MagicMock()

        self.trigger_reponse._trigger_response()
        self.trigger_reponse.before_buffer.assert_has_calls(
            [call.copy_to(ANY, seconds=SECONDS_BEFORE), call.clear()])
        self.trigger_reponse.streamer.assert_has_calls([call.split_recording(ANY),
                                                        call.wait_recording(
                                                            SECONDS_AFTER),
                                                        call.split_recording(self.trigger_reponse.before_buffer)])


if __name__ == '__main__':
    unittest.main()



# Globals config keys
TRUSTED_FACES = "trusted_faces"
MOTION_DETECTOR_THRESHOLD = "motion_detector_threshold"
MOTION_DETECTOR_SENSITIVITY = "motion_detector_sensitivity"
TRIGGER_COOLDOWN_TIME = "trigger_cooldown_time"

# Global Default
TRUSTED_FACES_DEFAULT = []
MOTION_DETECTOR_THRESHOLD_DEFAULT = 15  # Magnitude of motion vector
MOTION_DETECTOR_SENSITIVITY_DEFAULT = 25  # Number of motion vectors
TRIGGER_COOLDOWN_TIME_DEFAULT = 5

GLOBALS = {
    TRUSTED_FACES: TRUSTED_FACES_DEFAULT,
    MOTION_DETECTOR_THRESHOLD: MOTION_DETECTOR_THRESHOLD_DEFAULT,
    MOTION_DETECTOR_SENSITIVITY: MOTION_DETECTOR_SENSITIVITY_DEFAULT,
    TRIGGER_COOLDOWN_TIME: TRIGGER_COOLDOWN_TIME_DEFAULT
}



import json
from time import sleep
from .base import BaseMode
from outputs.server_socket_wrap import SocketWrap, SOCKET_TYPES
from streamers.streamer import Streamer

DEFAULT_FORMAT = "h264"


class NetworkStreamingMode(BaseMode):
    def __init__(self, config, trigger_responses):
        super().__init__(config, trigger_responses)
        self.name = "network_streaming"
        self.format = config["format"] if "format" in config else DEFAULT_FORMAT
        self.socket_map = {}
        self.has_hostsfile = False
        if "hosts" in config:
            self.hosts = config["hosts"]
        elif "hostsfile" in config:
            self.has_hostsfile = True
            self.hostsfile = config["hostsfile"] + (".%s" % self.format)
            self.hosts = self._read_hosts_file(retry=True)
        else:
            print("Error: No 'hosts' nor 'hostsfile' in config.")
            self.ready = False
            return

        prepend_size = config["prepend_size"] if "prepend_size" in config else False
        recording_options = config["recording_options"] if "recording_options" in config else {}
        self.streamer = Streamer(self.format, prepend_size=prepend_size, recording_options=recording_options)

    def _read_hosts_file(self, retry=False, retries=5, timeout=2):
        retry_count = 0
        while retry_count <= retries:
            try:
                with open(self.hostsfile, "r") as hostsfile:
                    return json.loads(hostsfile.read())
            except:
                if not retry:
                    return {}
                retry_count += 1
                sleep(timeout)
        return {}

    def update_socket_map(self):
        new_hosts_map = self._read_hosts_file()
        current_hosts = self.hosts.keys()
        new_hosts = new_hosts_map.keys()
        hosts_to_add = new_hosts - current_hosts
        hosts_to_delete = current_hosts - new_hosts

        for host in hosts_to_add:
            host_def = new_hosts_map[host]
            self._parse_and_add_host(host, host_def)

        for host in hosts_to_delete:
            self._remove_host(host)

        self.hosts = new_hosts_map

    def _parse_and_add_host(self, host, host_def):
        port = int(host_def["port"])
        protocol = host_def["protocol"]
        greeting = host_def["greeting"] if "greeting" in host_def.keys() else None
        if protocol not in SOCKET_TYPES:
            print("Protocol '%s' is not supported. Choose udp or tcp.")

        sock = SocketWrap(host, port, protocol, greeting)
        self.socket_map[host] = sock

    def _remove_host(self, host):
        if self.streamer.output.has_output(host):
            self.streamer.output.remove_output(host)

        if host in self.socket_map:
            self.socket_map[host].close()
            del self.socket_map[host]

    def pre_routine(self):
        for host, host_def in self.hosts.items():
            self._parse_and_add_host(host, host_def)

    def routine(self):
        # Update according to hostsfile in case it exists
        if self.has_hostsfile:
            self.update_socket_map()

        # Reconnection Routine
        for host, sock in self.socket_map.items():
            if not sock.is_connected():
                print("%s is not connected. Trying to reconnect." % host)
                if sock.connect():
                    if not self.streamer.output.has_output(host):
                        self.streamer.output.add_output(host, sock)
                else:
                    if self.streamer.output.has_output(host):
                        self.streamer.output.remove_output(host)
        sleep(2)

    def _cleanup(self):
        for host, sock in self.socket_map.items():
            if self.streamer.output.has_output(host):
                self.streamer.output.remove_output(host)
            sock.close()
        self.hosts_map = {}
        self.socket_map = {}



"""
This is the OutputHolder class.

This class helps streamers reuse camera ports for the same format and size combination
by passing the captured content through the various output analyzers
"""

from time import sleep
from collections import deque
from threading import Thread, Event, Lock, active_count, enumerate
from concurrent.futures import ThreadPoolExecutor


class OutputHolder:
    def __init__(self, outputs={}):
        self.outputs = outputs.copy()
        self.outputs_to_add = {}
        self.writable = False
        self.output_lock = Lock()
        self.writer_thread = None
        self.pool = ThreadPoolExecutor(max_workers=10)

    def has_outputs(self):
        return len(self.outputs) > 0

    def has_output(self, output_id):
        return output_id in self.outputs.keys()

    def get_output(self, output_id):
        return self.outputs[output_id]

    def add_output(self, output_id, output):
        with self.output_lock:
            self.outputs[output_id] = output

    def prepare_new_output(self, output_id, output):
        self.outputs_to_add[output_id] = output

    def remove_output(self, output_id):
        with self.output_lock:
            del self.outputs[output_id]

    def shutdown(self, wait=True):
        self.pool.shutdown(wait)


class WriterOutputHolder(Thread, OutputHolder):
    def __init__(self, outputs={}):
        Thread.__init__(self)
        OutputHolder.__init__(self, outputs)
        self.aux_buffer = deque()
        self.rebind_wait = Event()
        self.want_split = False
        self.outputs_to_split = {}

    def split_recording(self, output_id, output):
        self.want_split = True
        self.outputs_to_split[output_id] = output

    def write(self, buffer):
        """
        write method must return fast in order not to slow down the framerate
        """
        if self.writable:
            self.aux_buffer.append(buffer)

    def wait_for_rebind(self):
        self.rebind_wait.wait()
        self.rebind_wait.clear()

    def notify_split(self):
        self.aux_buffer.append(True)

    def run(self):
        self.writable = True
        with ThreadPoolExecutor(max_workers=10) as pool:
            while self.writable:
                try:
                    buffer = self.aux_buffer.popleft()

                    futures = []
                    with self.output_lock:
                        for oid, output in self.outputs.items():
                            # Using pool because instantiating Threads on every write is pretty slow
                            futures.append((oid, pool.submit(output.write, buffer)))
                        for oid, future in futures:
                            # Wait for result, if false remove the output
                            try:
                                if not future.result():
                                    del self.outputs[oid]
                            except Exception as e:
                                print("error writing to output %s %s:" % (oid, self.outputs[oid]), e)
                                del self.outputs[oid]

                    #  Check if its a split notifier
                    if type(buffer) is bool:
                        for out_id, out in self.outputs_to_add.items():
                            print("Changing output with id: %s" % out_id)
                            self.outputs[out_id] = out
                        self.outputs_to_add = {}
                        self.rebind_wait.set()
                        continue
                except IndexError:
                    sleep(0.05)
                except Exception as e:
                    print("Error writing frame to outputs:", e)
                    print("threads:\n", active_count(), enumerate())
                    self.stop()

    def stop(self):
        self.writable = False



import sys
sys.path.append("/picameleon")
from modes.network_streaming import NetworkStreamingMode
import unittest
from unittest.mock import Mock

mock_streamer = Mock()
mock_trigger_responses = [Mock() for _ in range(5)]

hosts_and_hostsfile_config = {
    "hosts": {
        "localhost": {}
    },
    "hostsfile": "some_file.json"
}

hostsfile_only_config = {
    "hostsfile": "some_file.json"
}

hostsfile_config = {
    "hosts": {
        "otherhost": {}
    }
}


def hosts_config_generator(hosts):
    host_config = {}
    for host in hosts:
        host_config[host] = {
            "port": 12345,
            "protocol": "tcp",
            "greeting": [0, 1, 0]
        }
    return host_config


# Just used to mock some functions that would need access to files
class MockNetworkStreamingMode(NetworkStreamingMode):
    def __init__(self, config, trigger_responses, return_value=None):
        self._return_value = return_value
        super().__init__(config, trigger_responses)

    def _read_hosts_file(self, retry=False, retries=5, timeout=2):
        return self._return_value


class TestNetworkStreamingMode(unittest.TestCase):

    def setUp(self):
        pass

    def test_sets_host_instead_of_hostfile(self):
        mode = MockNetworkStreamingMode(
            hosts_and_hostsfile_config, {}, return_value=hostsfile_config["hosts"])
        self.assertEqual(mode.hosts, hosts_and_hostsfile_config["hosts"])

    def test_sets_host_from_hostfile(self):
        mode = MockNetworkStreamingMode(
            hostsfile_only_config, {}, return_value=hostsfile_config["hosts"])
        self.assertEqual(mode.hosts, hostsfile_config["hosts"])

    def test_sets_mode_readiness_to_false_when_missing_hosts(self):
        mode = MockNetworkStreamingMode({}, {})
        self.assertFalse(mode.ready)

    def test_update_socket_map(self):
        first_hosts = hosts_config_generator(["1.1.1.1", "1.1.1.2"])
        hostsfile_config = hosts_config_generator(["1.1.1.3", "1.1.1.4"])
        mode = MockNetworkStreamingMode(
            hostsfile_only_config, {}, return_value=hostsfile_config)
        mode.hosts = first_hosts
        mode.update_socket_map()
        self.assertEqual(mode.hosts, hostsfile_config)
        self.assertEqual(mode.socket_map.keys(), mode.hosts.keys())

    def test_routine(self):
        socket_wrap1 = Mock()
        socket_wrap1.is_connected.return_value = False
        socket_wrap1.connect.return_value = True
        socket_wrap2 = Mock()
        socket_wrap2.is_connected.return_value = False
        socket_wrap2.connect.return_value = False
        config = {
            "hosts": {}
        }
        mode = MockNetworkStreamingMode(config, {})
        mode.socket_map = {
            "1.1.1.1": socket_wrap1,
            "1.1.1.2": socket_wrap2
        }
        mode.streamer.output.outputs = {"1.1.1.2": socket_wrap2}
        mode.routine()
        self.assertEqual(mode.streamer.output.outputs,
                         {"1.1.1.1": socket_wrap1})


if __name__ == '__main__':
    unittest.main()



import setuptools


def main():
    setuptools.setup(
        name='picameleon',
        packages=["picameleon", "picameleon.outputs"],
        include_package_data=True,
        platforms='ALL',
        version='0.0.2',
        license='LGPLv3',
        description='A python client or picameleon',
        author='André Esser',
        author_email='bepandre@hotmail.com',
        url='https://github.com/Esser50K/PiCameleon',
        download_url='https://github.com/Esser50K/PiCameleon/archive/refs/tags/picameleon-client-v0.0.2.tar.gz',
        keywords=['picamera', 'service', 'raspberrypi'],
        install_requires=[],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
        ],
    )


if __name__ == '__main__':
    main()



import os
import tempfile
import requests
from time import sleep, time
from datetime import datetime
from streamers.streamer import ROOT_PATH, Streamer
from utils.single_picamera import SinglePiCamera
from .base import BaseTriggerResponse
from threading import Thread
from concurrent.futures import ThreadPoolExecutor

HTTP_TRIGGER_RESPONSE_NAME = "http"


class HTTP(BaseTriggerResponse):
    def __init__(self, http_service_urls=None, **capture_options):
        if http_service_urls is None:
            http_service_urls = {}

        super().__init__(use_detrigger=True)
        self.trigger_name = HTTP_TRIGGER_RESPONSE_NAME
        self.capture_options = capture_options
        # the key is the url the value is a boolean if a picture should be sent with the request
        self.http_service_urls = http_service_urls

    def _detriggered(self):
        return self.detrigger_event.wait(0.1)

    def _open_and_send(self, url, image_path):
        files = {"file": open(image_path, "rb")}
        result = requests.post(url, files=files)
        files["file"].close()
        return result

    def _trigger_response(self, trigger_args={}):
        tmp_dir = tempfile.TemporaryDirectory()
        image_path = tmp_dir.name + '/http_' + str(int(time())) + '.jpg'
        try:
            Streamer.take_picture(image_path, format='jpeg', **self.capture_options)
            for url, send_photo in self.http_service_urls.items():
                result = self._open_and_send(
                    url, image_path) if send_photo else requests.post(url)

                if result.status_code != 200:
                    print("http request %s was not successful:" %
                          url, result.text)
        finally:
            tmp_dir.cleanup()



import sys
sys.path.append("/picameleon")
from mocks.mock_socket import MockSocket
from modes.network_trigger import NetowrkTriggerMode
import json
import unittest
import struct
from unittest.mock import Mock, call, patch

mode = None
mode_config = {
    "redis": True,
    "redis_addr": "localhost"
}


class TestNetworkServingMode(unittest.TestCase):
    def tearDown(self):
        global mode
        if mode:
            mode._cleanup()

    @patch("redis.Redis")
    @patch("socket.socket")
    def test_preroutine(self, mocked_redis, mocked_socket):
        global mode
        mode = NetowrkTriggerMode(mode_config, {})
        mode.pre_routine()
        self.assertIsNotNone(mode.server_socket)
        mode.server_socket.assert_has_calls(
            [call.bind(("0.0.0.0", 5556)), call.listen()])
        self.assertIsNotNone(mode.redis)
        self.assertIsNotNone(mode.redis_announcer)

    @patch("redis.Redis")
    @patch("socket.socket")
    def test_routine(self, mocked_redis, mocked_socket):
        global mode
        mode = NetowrkTriggerMode(mode_config, {})
        mode.trigger = Mock()
        mode.pre_routine()
        client_address = "localhost:1234"
        request = json.dumps({"format": "h264", "bitrate": "vhigh"})
        mock_request = MockSocket([struct.pack("<L", len(request)), request])
        mode.server_socket.accept = Mock(
            return_value=(mock_request, client_address))
        mode.routine()
        self.assertTrue(mode.is_triggered)
        mode.trigger.assert_called_once()

        # should only trigger once until it is detriggered
        mode.routine()
        mode.trigger.assert_called_once()

        mode.is_triggered = False
        mode.routine()
        self.assertEqual(mode.trigger.call_count, 2)

    @patch("redis.Redis")
    @patch("socket.socket")
    def test_cleanup(self, mocked_redis, mocked_socket):
        global mode
        mode = NetowrkTriggerMode(mode_config, {})
        mode.pre_routine()
        mode._cleanup()
        mode.server_socket.assert_has_calls([call.close()])
        self.assertIsNone(mode.redis_announcer)


if __name__ == '__main__':
    unittest.main()



import struct
import socket
from typing import Dict, Any
from json import dumps
from dataclasses import dataclass
from threading import Thread, Event, Lock
from ..outputs import WriterOutputHolder

DEFAULT_FRAME_FETCH_CHUNK_SIZE = 4096
DEFAULT_STREAM_SERVER_PORT = 5555
DEFAULT_TRIGGER_SERVER_PORT = 5556


@dataclass()
class StreamInfo:
    event: Event
    thread: Thread
    running: bool
    port: int = 0


class Client:
    def __init__(self, address):
        self.address = address
        self.current_streams: Dict[str, StreamInfo] = {}
        self.current_frames: Dict[str, bytes] = {}
        self.output_holders: Dict[str, WriterOutputHolder] = {}
        self.__lock = Lock()

    def get_next_frame(self, stream_id, timeout=None):
        if stream_id not in self.current_frames.keys() or stream_id not in self.current_streams.keys():
            raise Exception("no frames to fetch under stream with id %s" % stream_id)

        if self.current_streams[stream_id].event.wait(timeout):
            return self.current_frames[stream_id]

    def stop_stream(self, stream_id: str):
        with self.__lock:
            if stream_id in self.current_streams.keys():
                self.current_streams[stream_id].running = False
                if self.__is_receiver(stream_id):
                    self.__stop_receiver(stream_id)
                if stream_id in self.output_holders.keys():
                    self.output_holders[stream_id].stop()
                    self.output_holders[stream_id].shutdown()
                    del self.output_holders[stream_id]
        self.current_streams[stream_id].thread.join()

    def __is_receiver(self, stream_id: str):
        return self.current_streams[stream_id].port != 0

    def __stop_receiver(self, stream_id: str):
        if stream_id in self.current_streams.keys():
            self.current_streams[stream_id].running = False
            try:
                sock = socket.socket()
                sock.connect(('0.0.0.0', self.current_streams[stream_id].port))
            except socket.error as e:
                print("server wasn't listening anymore:", e)

    def serve_stream_receiver(self, stream_id: str, port: int):
        with self.__lock:
            event = Event()
            thread = Thread(target=self.__read_stream, args=(stream_id, port))
            self.current_streams[stream_id] = StreamInfo(event, thread, True, port)
            self.current_frames[stream_id] = b''
            thread.start()

    def __read_stream(self, stream_id: str, port: int):
        server_socket = socket.socket()
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(('0.0.0.0', port))
        server_socket.listen(0)
        connection = server_socket.accept()[0]
        server_socket.close()

        try:
            while self.current_streams[stream_id].running:
                frame_size = struct.unpack('<L', connection.recv(4))[0]
                if frame_size == 0:
                    break

                frame = b''
                while len(frame) != frame_size:
                    frame += connection.recv(frame_size - len(frame))
                self.current_frames[stream_id] = frame
                if stream_id in self.output_holders.keys():
                    self.output_holders[stream_id].write(frame)
                self.current_streams[stream_id].event.set()
        except Exception as e:
            print("error reading stream:", e)
        finally:
            connection.close()
            self.current_streams[stream_id].event.set()
            with self.__lock:
                del self.current_streams[stream_id]
                del self.current_frames[stream_id]

    def fetch_stream(self,
                     stream_id: str,
                     stream_format: str,
                     port=DEFAULT_STREAM_SERVER_PORT,
                     resize: list = None,
                     bitrate: int = None,
                     frame_size=DEFAULT_FRAME_FETCH_CHUNK_SIZE,
                     size_prepended=True) -> None:
        with self.__lock:
            if stream_id in self.current_streams.keys():
                raise Exception("stream with id %s is still active" % stream_id)

            event = Event()
            thread = Thread(target=self.__fetch_stream,
                            args=(stream_id, stream_format, port, resize, bitrate, frame_size, size_prepended))
            self.current_streams[stream_id] = StreamInfo(event, thread, True)
            self.current_frames[stream_id] = b''
            thread.start()

    def write_stream_into(self, stream_id, output_id, output):
        with self.__lock:
            if stream_id not in self.current_streams.keys():
                raise Exception("stream with id %s is not active" % stream_id)

            if stream_id not in self.output_holders.keys():
                self.output_holders[stream_id] = WriterOutputHolder()
                self.output_holders[stream_id].start()

            output = parse_output(output)
            self.output_holders[stream_id].add_output(output_id, output)

    def remove_output_from_stream(self, stream_id, output_id):
        with self.__lock:
            if stream_id not in self.output_holders.keys():
                return

            self.output_holders[stream_id].remove_output(output_id)
            if len(self.output_holders[stream_id].outputs) == 0:
                self.output_holders[stream_id].stop()
                self.output_holders[stream_id].shutdown()

    def __fetch_stream(self,
                       stream_id: str,
                       stream_format: str,
                       port=DEFAULT_STREAM_SERVER_PORT,
                       resize: tuple = None,
                       bitrate: int = None,
                       frame_size=DEFAULT_FRAME_FETCH_CHUNK_SIZE,
                       size_prepended=True) -> None:
        sock = socket.socket()
        sock.settimeout(1)
        try:
            sock.connect((self.address, port))
        except Exception as e:
            print("failed to connect to picameleon to fetch stream at %s:%s: %s" % (self.address, port, e))
            sock.close()
            self.current_streams[stream_id].event.set()
            with self.__lock:
                del self.current_streams[stream_id]
                del self.current_frames[stream_id]

        try:
            stream_request = {"format": stream_format}
            if bitrate:
                stream_request["bitrate"] = bitrate
            if resize:
                stream_request["resize"] = resize

            stream_request = dumps(stream_request).encode("utf-8")
            request_size = len(stream_request)
            sock.send(struct.pack('<L', request_size))
            sock.sendall(stream_request)

            while self.current_streams[stream_id].running:
                self.current_streams[stream_id].event.clear()
                if size_prepended:
                    frame_size = struct.unpack('<L', sock.recv(4))[0]
                    if frame_size == 0:
                        break

                frame = b''
                while len(frame) != frame_size:
                    frame += sock.recv(frame_size - len(frame))

                self.current_frames[stream_id] = frame
                if stream_id in self.output_holders.keys():
                    self.output_holders[stream_id].write(frame)
                self.current_streams[stream_id].event.set()
        except Exception as e:
            print("failed to read frame from picameleon stream at %s:%s: %s" % (self.address, port, e))
        finally:
            sock.close()
            self.current_streams[stream_id].event.set()
            with self.__lock:
                del self.current_streams[stream_id]
                del self.current_frames[stream_id]

    def trigger(self, port=DEFAULT_TRIGGER_SERVER_PORT):
        sock = socket.socket()
        sock.connect((self.address, port))
        sock.close()

    def shutdown(self):
        for stream_id in list(self.current_streams.keys()):
            self.stop_stream(stream_id)


def parse_output(output: Any):
    if type(output) is str:
        return open(output, "w")

    if getattr(output, "write", None):
        return output

    raise Exception("%s object is not recognized as writable" % output)



"""
The Motion Detection class.
"""

import numpy
from threading import Event
from utils.consts import GLOBALS, MOTION_DETECTOR_THRESHOLD, MOTION_DETECTOR_SENSITIVITY

OUTPUT_ID = "motion_detector"


class MotionDetector:
    def __init__(self, threshold=GLOBALS[MOTION_DETECTOR_THRESHOLD], sensitivity=GLOBALS[MOTION_DETECTOR_SENSITIVITY]):
        self.threshold = threshold
        self.sensitivity = sensitivity
        self.motion_detected = Event()
        self.save = True
        self.motion_vectors_list = []
        self.start = 0

    # Jsonify numpy arrays
    # https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
    def analyze(self, motion_vectors):
        self.motion_detected.clear()
        motion_vectors = numpy.sqrt(
            numpy.square(motion_vectors['x'].astype(numpy.float)) +
            numpy.square(motion_vectors['y'].astype(numpy.float))
        ).clip(0, 255).astype(numpy.uint8)
        if (motion_vectors >= self.threshold).sum() > self.sensitivity:
            self.motion_detected.set()

    def calibrate(self, no_motion_seconds=5, max_resets=5):
        reset_count = 0
        while self.motion_detected.wait(no_motion_seconds):
            reset_count += 1
            if reset_count > max_resets:
                return False
        return True

    def wait_for_motion(self, timeout=None):
        return self.motion_detected.wait(timeout)



import io
import struct
from threading import Thread, Event
from .base import BaseStreamer

SPLITTER_STRINGS = {
    "mjpeg": b'\xff\xd8',
    "h264": b'\x00\x00\x00\x01'
}


class SplitFrameStreamer(BaseStreamer):
    def __init__(self, port, format, prepend_size=False, options=None):
        if options is None:
            options = {}

        super().__init__(port, format, options)
        self.event = Event()
        self.prepend_size = prepend_size
        self.stream = io.BytesIO()
        self.last_frame = None
        self.streamer_thread = None
        self.splitter_string = SPLITTER_STRINGS[format] if format in SPLITTER_STRINGS.keys() else None

    def _setup_streamer(self):
        self.sub_output = self
        if self.output:
            self.streamer_thread = Thread(target=self.stream_frames)
            self.streamer_thread.start()

    def get_last_frame(self):
        return self.last_frame

    def get_next_frame(self, timeout=None):
        if self.event.wait(timeout):
            return self.last_frame

    def stream_frames(self):
        while self.is_running:
            try:
                frame = self.get_next_frame(timeout=1)
                if frame is None:
                    continue

                self.event.clear()
                if self.prepend_size:
                    frame = struct.pack('<L', len(frame)) + frame
                self.output.write(frame)
            except Exception as e:
                print("Error writing next frame:", e)

    def can_write(self, buf):
        # check if buffer starts with splitter string if it exists
        # if it doesn't exist then just check the stream is not empty
        return ((self.splitter_string is not None and \
                buf.startswith(self.splitter_string)) or \
               (self.splitter_string is None)) and \
               self.stream.tell() > 0


    def write(self, buf):
        if self.can_write(buf):
            self.stream.seek(0)
            self.last_frame = self.stream.read()
            self.event.set()
            self.stream.seek(0)
            self.stream.truncate()

        self.stream.write(buf)

    def _teardown_streamer(self):
        if self.streamer_thread:
            self.streamer_thread.join()
            self.streamer_thread = None



import sys
sys.path.append("/picameleon")
from mocks.mock_socket import MockSocket
from modes.network_serving import NetworkServingMode
import json
import unittest
import struct
from unittest.mock import Mock, call, patch

mode = None
mode_config = {
    "redis": True,
    "redis_addr": "localhost"
}


class TestNetworkTriggerMode(unittest.TestCase):
    def tearDown(self):
        if mode:
            mode._cleanup()

    @patch("redis.Redis")
    @patch("socket.socket")
    def test_preroutine(self, mocked_redis, mocked_socket):
        global mode
        mode = NetworkServingMode(mode_config, {})
        mode.pre_routine()
        self.assertIsNotNone(mode.server_socket)
        mode.server_socket.assert_has_calls(
            [call.bind(("0.0.0.0", 5555)), call.listen()])
        self.assertIsNotNone(mode.redis)
        self.assertIsNotNone(mode.redis_announcer)

    @patch("redis.Redis")
    @patch("socket.socket")
    def test_routine(self, mocked_redis, mocked_socket):
        global mode
        mode = NetworkServingMode(mode_config, {})
        mode.pre_routine()
        client_address = "localhost:1234"
        request = json.dumps({"format": "h264", "bitrate": 1_000_000})
        mode.server_socket.accept = Mock(return_value=(MockSocket(
            [struct.pack("<L", len(request)), request]), client_address))
        mode.routine()
        self.assertEqual(len(mode.socket_map), 1)
        self.assertTrue(client_address in mode.socket_map.keys())
        self.assertTrue(client_address in mode.socket_to_stream.keys())
        self.assertTrue(
            mode.streamer_map["bitrate_1000000format_h264"].output.has_output(client_address))
        mode.socket_map[client_address].close()
        self.assertEqual(len(mode.socket_map), 0)

    @patch("redis.Redis")
    @patch("socket.socket")
    def test_cleanup(self, mocked_redis, mocked_socket):
        global mode
        mode = NetworkServingMode(mode_config, {})
        mode.pre_routine()
        client_socket_mock = Mock()
        mode.socket_map = {
            "localhost:1234": client_socket_mock,
            "localhost:1235": client_socket_mock
        }
        mode._cleanup()
        mode.server_socket.assert_has_calls([call.close()])
        client_socket_mock.assert_has_calls([call.close()])
        self.assertIsNone(mode.redis_announcer)


if __name__ == '__main__':
    unittest.main()



import os
from time import sleep, time
from datetime import datetime
from utils.single_picamera import SinglePiCamera
from streamers.streamer import Streamer, ROOT_PATH
from .base import BaseTriggerResponse
from picamerax import PiCameraCircularIO
from threading import Thread


FILE_BUFFER = 1048576


class RecordToFile(BaseTriggerResponse):
    def __init__(self, format, destination_path,
                 record_time_before_trigger, record_time_after_trigger,
                 recording_options={}, use_detrigger=True, file_buffer=FILE_BUFFER):
        super().__init__(use_detrigger=use_detrigger)
        self.format = format
        self.recording_options = recording_options
        self.seconds_before = int(record_time_before_trigger)
        self.seconds_after = int(record_time_after_trigger)
        self.destination_path = destination_path
        self.file_buffer = file_buffer
        self.before_buffer = None
        self.streamer = None
        self.is_triggered = False

    def _initialize_trigger_response(self):
        self.streamer = Streamer(
            self.format, split_frames=False, recording_options=self.recording_options)
        self.before_buffer = PiCameraCircularIO(
            SinglePiCamera(), seconds=self.seconds_before, splitter_port=self.streamer.port)
        self.streamer.output = self.before_buffer
        self.streamer.start()

    def _trigger_response(self, trigger_args=None):
        # Mutliple modes can use this trigger response.
        # Don't record for more than 1 trigger response
        if not self.is_triggered:
            self.is_triggered = True
            if not os.path.exists(self.destination_path):
                os.makedirs(self.destination_path)

            now = datetime.now()
            path = "{d}/{t}.{f}".format(d=self.destination_path,
                                        t=now.strftime("%Y-%m-%d_%H-%M"), f=self.format)
            outputfile = open(path, mode="wb", buffering=self.file_buffer)
            self.before_buffer.copy_to(outputfile, seconds=self.seconds_before)
            self.streamer.split_recording(outputfile)

            # Wait until triggered to stop
            if self.use_detrigger:
                self.detrigger_event.wait()
                self.detrigger_event.clear()

            # Still record X seconds after detrigger
            self.streamer.wait_recording(self.seconds_after)
            self.before_buffer.clear()
            self.streamer.split_recording(self.before_buffer)
            outputfile.close()
            self.is_triggered = False

    def _cleanup_trigger_response(self):
        self.streamer.stop()
        self.before_buffer.clear()
